/*
**	ImportPanel.c
**
**	Editing panel for phone unit settings
**
**	Copyright  1990-1993 by Olaf `Olsen' Barthel & MXM
**		All Rights Reserved
*/

#include "termGlobal.h"

enum	{	GAD_LIST,GAD_REPLACE,GAD_APPEND,GAD_CANCEL };

STATIC struct Hook ListViewHook;

	/* ListViewRender():
	 *
	 *	This callback routine is responsible for rendering
	 *	the single listview entries.
	 */

STATIC ULONG __saveds __asm
ListViewRender(register __a0 struct Hook *Hook,register __a1 struct LVDrawMsg *Msg,register __a2 struct Node *Node)
{
		/* We only know how to redraw lines. */

	if(Msg -> lvdm_MethodID == LV_DRAW)
	{
		struct RastPort	*RPort	= Msg -> lvdm_RastPort;
		LONG		 Left	= Msg -> lvdm_Bounds . MinX,
				 Top	= Msg -> lvdm_Bounds . MinY,
				 Width	= Msg -> lvdm_Bounds . MaxX - Msg -> lvdm_Bounds . MinX + 1,
				 Height	= Msg -> lvdm_Bounds . MaxY - Msg -> lvdm_Bounds . MinY + 1,
				 FgPen,BgPen,
				 Len,LabelLen;
		STRPTR		 String;

		String = &Node -> ln_Name[6];

			/* Determine the rendering pens. */

		switch(Msg -> lvdm_State)
		{
			case LVR_SELECTEDDISABLED:
			case LVR_NORMALDISABLED:
			case LVR_NORMAL:

				FgPen = Msg -> lvdm_DrawInfo -> dri_Pens[TEXTPEN];
				BgPen = Msg -> lvdm_DrawInfo -> dri_Pens[BACKGROUNDPEN];
				break;

			case LVR_SELECTED:

				FgPen = Msg -> lvdm_DrawInfo -> dri_Pens[FILLTEXTPEN];
				BgPen = Msg -> lvdm_DrawInfo -> dri_Pens[FILLPEN];
				break;
		}

		SetAPen(RPort,BgPen);
		RectFill(RPort,Left,Top,Left + 1,Top + Height - 1);

		Left	+= 2;
		Width	-= 2;

			/* Set the rendering pens. */

		SetDrMd(RPort,JAM2);
		SetAPen(RPort,FgPen);
		SetBPen(RPort,BgPen);

			/* Determine length of vanilla name. */

		LabelLen = strlen(String);

			/* Try to make it fit. */

		while(LabelLen > 0 && (Len = TextLength(RPort,String,LabelLen)) > Width)
			LabelLen--;

			/* Print the vanilla name if possible. */

		if(LabelLen)
		{
			Move(RPort,Left,Top + RPort -> Font -> tf_Baseline);
			Text(RPort,String,LabelLen);

			Left	+= Len;
			Width	-= Len;
		}

			/* Fill the area to follow the vanilla name. */

		if(Width > 0)
		{
			SetAPen(RPort,BgPen);
			RectFill(RPort,Left,Top,Left + Width - 1,Top + Height - 1);
		}

			/* If the item happens to be disabled, draw the cross-hatch
			 * pattern across it.
			 */

		if(Msg -> lvdm_State == LVR_SELECTEDDISABLED || Msg -> lvdm_State == LVR_NORMALDISABLED)
		{
			SetDrMd(RPort,JAM1);

			SetAPen(RPort,Msg -> lvdm_DrawInfo -> dri_Pens[BLOCKPEN]);

			SetAfPt(RPort,(UWORD *)&Ghosting,1);

			RectFill(RPort,Msg -> lvdm_Bounds . MinX,Msg -> lvdm_Bounds . MinY,Msg -> lvdm_Bounds . MaxX,Msg -> lvdm_Bounds . MaxY);

			SetAfPt(RPort,NULL,0);
		}

		return(LVCB_OK);
	}
	else
		return(LVCB_UNKNOWN);
}

STATIC struct Gadget *
CreateAllGadgets(struct List *PhoneList,struct Gadget **GadgetArray,struct Gadget **GadgetList)
{
	struct Gadget		*Gadget;
	struct NewGadget	 NewGadget;
	UWORD			 Counter = 0;

	SZ_SizeSetup(Window -> WScreen,&UserFont,TRUE);

	memset(&NewGadget,0,sizeof(struct NewGadget));

	if(Gadget = CreateContext(GadgetList))
	{
		UWORD Width,ButtonWidth;

		SZ_ResetMaxWidth();

		SZ_UpdateMaxWidth(BUTTON_KIND,LocaleString(MSG_IMPORTPANEL_REPLACE_RATES_GAD),0,NULL);
		SZ_UpdateMaxWidth(BUTTON_KIND,LocaleString(MSG_IMPORTPANEL_APPEND_RATES_GAD),0,NULL);
		SZ_UpdateMaxWidth(BUTTON_KIND,LocaleString(MSG_GLOBAL_CANCEL_GAD),0,NULL);

		ButtonWidth = SZ_ResetMaxWidth();

		Width = SZ_Width(LISTVIEW_KIND,NULL,35,NULL);

		if(3 * ButtonWidth + 2 * InterWidth > Width)
			SZ_SetWidth(3 * ButtonWidth + 2 * InterWidth);
		else
			SZ_SetWidth(Width);

		NewGadget . ng_GadgetText	= LocaleString(MSG_GLOBAL_NAME_LIST_GAD);
		NewGadget . ng_GadgetID		= Counter;
		NewGadget . ng_Flags		= PLACETEXT_ABOVE|NG_HIGHLABEL;

		ListViewHook . h_Entry = (LONG (*)())ListViewRender;

		GadgetArray[Counter++] = Gadget = CreateGadget(LISTVIEW_KIND,Gadget,&NewGadget,
			SZ_Adjust,		TRUE,
			SZ_AutoWidth,		TRUE,
			SZ_Lines,		8,

			GTLV_CallBack,		&ListViewHook,
			GTLV_Labels,		PhoneList,
			GTLV_ShowSelected,	NULL,
		TAG_DONE);

		SZ_SetWidth(ButtonWidth);

		NewGadget . ng_GadgetText	= LocaleString(MSG_IMPORTPANEL_REPLACE_RATES_GAD);
		NewGadget . ng_GadgetID		= Counter;
		NewGadget . ng_Flags		= 0;

		GadgetArray[Counter++] = Gadget = CreateGadget(BUTTON_KIND,Gadget,&NewGadget,
			SZ_Adjust,	TRUE,
			SZ_AutoWidth,	TRUE,
			SZ_GroupCount,	3,

			GA_Disabled,	TRUE,
		TAG_DONE);

		NewGadget . ng_GadgetText	= LocaleString(MSG_IMPORTPANEL_APPEND_RATES_GAD);
		NewGadget . ng_GadgetID		= Counter;

		GadgetArray[Counter++] = Gadget = CreateGadget(BUTTON_KIND,Gadget,&NewGadget,
			SZ_Adjust,	TRUE,
			SZ_AutoWidth,	TRUE,
			SZ_GroupNext,	TRUE,

			GA_Disabled,	TRUE,
		TAG_DONE);

		NewGadget . ng_GadgetText	= LocaleString(MSG_GLOBAL_CANCEL_GAD);
		NewGadget . ng_GadgetID		= Counter;

		GadgetArray[Counter] = Gadget = CreateGadget(BUTTON_KIND,Gadget,&NewGadget,
			SZ_Adjust,	TRUE,
			SZ_AutoWidth,	TRUE,
			SZ_GroupNext,	TRUE,
		TAG_DONE);
	}

	return(Gadget);
}

BYTE
ImportPanel(struct PhoneEntry *Entry,struct List *PhoneList)
{
	struct Window	*PanelWindow;
	struct Gadget	*GadgetList;
	struct Gadget	*GadgetArray[GAD_CANCEL + 1];
	BYTE		 MadeChanges = FALSE;

	if(CreateAllGadgets(PhoneList,GadgetArray,&GadgetList))
	{
		if(PanelWindow = OpenWindowTags(NULL,
			WA_Left,	GetScreenLeft(Window) + (GetScreenWidth(Window) - SZ_GetWindowWidth()) / 2,
			WA_Top,		GetScreenTop(Window) + (GetScreenHeight(Window) - SZ_GetWindowHeight())	/ 2,
			WA_Width,	SZ_GetWindowWidth(),
			WA_Height,	SZ_GetWindowHeight(),

			WA_Activate,	TRUE,
			WA_DragBar,	TRUE,
			WA_DepthGadget,	TRUE,
			WA_RMBTrap,	TRUE,
			WA_DepthGadget,	TRUE,
			WA_CloseGadget,	TRUE,
			WA_CustomScreen,Window -> WScreen,
			WA_IDCMP,	IDCMP_CLOSEWINDOW | IDCMP_VANILLAKEY | IDCMP_RAWKEY | BUTTONIDCMP | LISTVIEWIDCMP,
			WA_Title,	LocaleString(MSG_IMPORTPANEL_IMPORT_RATES_TXT),
		TAG_DONE))
		{
			struct IntuiMessage	*Massage;
			ULONG			 IClass,Code;
			struct Gadget		*Gadget;
			BYTE			 Terminated = FALSE;
			LONG			 Selected = ~0;

			GuideContext(CONTEXT_IMPORT);

			PushWindow(PanelWindow);

			AddGList(PanelWindow,GadgetList,(UWORD)-1,(UWORD)-1,NULL);
			RefreshGList(GadgetList,PanelWindow,NULL,(UWORD)-1);
			GT_RefreshWindow(PanelWindow,NULL);

			ClrSignal(SIG_BREAK);

			while(!Terminated)
			{
				if(Wait(PORTMASK(PanelWindow -> UserPort) | SIG_BREAK) & SIG_BREAK)
					break;

				while(!Terminated && (Massage = (struct IntuiMessage *)GT_GetIMsg(PanelWindow -> UserPort)))
				{
					IClass	= Massage -> Class;
					Code	= Massage -> Code;
					Gadget	= (struct Gadget *)Massage -> IAddress;

					GT_ReplyIMsg(Massage);

					KeySelect(GadgetArray,GAD_CANCEL,Code,PanelWindow,&Gadget,&IClass,&Code);

					if(IClass == IDCMP_CLOSEWINDOW)
						Terminated = TRUE;

					if(IClass == IDCMP_GADGETUP)
					{
						switch(Gadget -> GadgetID)
						{
							case GAD_LIST:

								if(Entry != Phonebook[Code])
								{
									if(Selected == ~0)
									{
										GT_SetGadgetAttrs(GadgetArray[GAD_REPLACE],PanelWindow,NULL,
											GA_Disabled,	FALSE,
										TAG_DONE);

										GT_SetGadgetAttrs(GadgetArray[GAD_APPEND],PanelWindow,NULL,
											GA_Disabled,	FALSE,
										TAG_DONE);
									}

									Selected = Code;
								}
								else
								{
									if(Selected != ~0)
									{
										GT_SetGadgetAttrs(GadgetArray[GAD_REPLACE],PanelWindow,NULL,
											GA_Disabled,	TRUE,
										TAG_DONE);

										GT_SetGadgetAttrs(GadgetArray[GAD_APPEND],PanelWindow,NULL,
											GA_Disabled,	TRUE,
										TAG_DONE);

										Selected = ~0;
									}
								}

								break;

							case GAD_APPEND:
							case GAD_REPLACE:

								if(Gadget -> GadgetID == GAD_REPLACE)
								{
									FreeTimeDateList((struct List *)&Entry -> TimeDateList);

									CopyTimeDateList((struct List *)&Phonebook[Selected] -> TimeDateList,(struct List *)&Entry -> TimeDateList,FALSE);
								}
								else
									CopyTimeDateList((struct List *)&Phonebook[Selected] -> TimeDateList,(struct List *)&Entry -> TimeDateList,TRUE);

								Terminated = TRUE;

								MadeChanges = TRUE;

								break;

							case GAD_CANCEL:

								Terminated = TRUE;

								break;
						}
					}
				}
			}

			RemoveGList(PanelWindow,GadgetList,(UWORD)-1);

			PopWindow();

			CloseWindow(PanelWindow);
		}

		FreeGadgets(GadgetList);
	}

	return(MadeChanges);
}
